@@ -22,17 +22,16 @@
 import sys
 import warnings
 
-from gi.repository import GObject
-from .._ossighelper import wakeup_on_signal, register_sigint_fallback
+from gi.repository import GObject, GLib
+from .._ossighelper import register_sigint_fallback, get_event_loop
 from .._gtktemplate import Template, _extract_handler_and_args
 from ..overrides import (override, strip_boolean_result, deprecated_init,
                          wrap_list_store_sort_func)
 from ..module import get_introspection_module
-from gi import PyGIDeprecationWarning
+from gi import PyGIDeprecationWarning, get_option
 
 
 Gtk = get_introspection_module('Gtk')
-GTK2 = Gtk._version == '2.0'
 GTK3 = Gtk._version == '3.0'
 GTK4 = Gtk._version == '4.0'
 
@@ -46,16 +45,6 @@ __all__.append('Template')
 _extract_handler_and_args = _extract_handler_and_args
 __all__.append('_extract_handler_and_args')
 
-if GTK2:
-    warn_msg = "You have imported the Gtk 2.0 module.  Because Gtk 2.0 \
-was not designed for use with introspection some of the \
-interfaces and API will fail.  As such this is not supported \
-by the pygobject development team and we encourage you to \
-port your app to Gtk 3 or greater. PyGTK is the recomended \
-python module to use with Gtk 2.0"
-
-    warnings.warn(warn_msg, RuntimeWarning)
-
 
 class PyGTKDeprecationWarning(PyGIDeprecationWarning):
     pass
@@ -64,21 +53,21 @@ class PyGTKDeprecationWarning(PyGIDeprecationWarning):
 __all__.append('PyGTKDeprecationWarning')
 
 
-def _construct_target_list(targets):
-    """Create a list of TargetEntry items from a list of tuples in the form (target, flags, info)
+if GTK3:
+    def _construct_target_list(targets):
+        """Create a list of TargetEntry items from a list of tuples in the form (target, flags, info)
 
-    The list can also contain existing TargetEntry items in which case the existing entry
-    is re-used in the return list.
-    """
-    target_entries = []
-    for entry in targets:
-        if not isinstance(entry, Gtk.TargetEntry):
-            entry = Gtk.TargetEntry.new(*entry)
-        target_entries.append(entry)
-    return target_entries
+        The list can also contain existing TargetEntry items in which case the existing entry
+        is re-used in the return list.
+        """
+        target_entries = []
+        for entry in targets:
+            if not isinstance(entry, Gtk.TargetEntry):
+                entry = Gtk.TargetEntry.new(*entry)
+            target_entries.append(entry)
+        return target_entries
 
-
-__all__.append('_construct_target_list')
+    __all__.append('_construct_target_list')
 
 
 def _builder_connect_callback(builder, gobj, signal_name, handler_name, connect_obj, flags, obj_or_map):
@@ -122,24 +111,24 @@ class Widget(Gtk.Widget):
                 yield child
                 child = child.get_next_sibling()
 
-    if GTK2 or GTK3:
+    if GTK3:
         def freeze_child_notify(self):
             super(Widget, self).freeze_child_notify()
             return _FreezeNotifyManager(self)
 
-    if GTK2 or GTK3:
+    if GTK3:
         def drag_dest_set_target_list(self, target_list):
             if (target_list is not None) and (not isinstance(target_list, Gtk.TargetList)):
                 target_list = Gtk.TargetList.new(_construct_target_list(target_list))
             super(Widget, self).drag_dest_set_target_list(target_list)
 
-    if GTK2 or GTK3:
+    if GTK3:
         def drag_source_set_target_list(self, target_list):
             if (target_list is not None) and (not isinstance(target_list, Gtk.TargetList)):
                 target_list = Gtk.TargetList.new(_construct_target_list(target_list))
             super(Widget, self).drag_source_set_target_list(target_list)
 
-    if GTK2 or GTK3:
+    if GTK3:
         def style_get_property(self, property_name, value=None):
             if value is None:
                 prop = self.find_style_property(property_name)
@@ -156,7 +145,7 @@ Widget = override(Widget)
 __all__.append('Widget')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Container(Gtk.Container, Widget):
 
         def __len__(self):
@@ -171,9 +160,6 @@ if GTK2 or GTK3:
         def __bool__(self):
             return True
 
-        # alias for Python 2.x object protocol
-        __nonzero__ = __bool__
-
         def child_get_property(self, child, property_name, value=None):
             if value is None:
                 prop = self.find_child_property(property_name)
@@ -215,7 +201,7 @@ Editable = override(Editable)
 __all__.append("Editable")
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Action(Gtk.Action):
         __init__ = deprecated_init(Gtk.Action.__init__,
                                    arg_names=('name', 'label', 'tooltip', 'stock_id'),
@@ -417,7 +403,7 @@ ComboBox = override(ComboBox)
 __all__.append('ComboBox')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Box(Gtk.Box):
         __init__ = deprecated_init(Gtk.Box.__init__,
                                    arg_names=('homogeneous', 'spacing'),
@@ -427,7 +413,7 @@ if GTK2 or GTK3:
     __all__.append('Box')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class SizeGroup(Gtk.SizeGroup):
         __init__ = deprecated_init(Gtk.SizeGroup.__init__,
                                    arg_names=('mode',),
@@ -438,7 +424,7 @@ if GTK2 or GTK3:
     __all__.append('SizeGroup')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class MenuItem(Gtk.MenuItem):
         __init__ = deprecated_init(Gtk.MenuItem.__init__,
                                    arg_names=('label',),
@@ -515,7 +501,7 @@ else:
 
 class Window(Gtk.Window):
     def __init__(self, *args, **kwargs):
-        if not initialized:
+        if not initialized and get_option('legacy_autoinit'):  # TODO: and not GTK5
             raise RuntimeError(
                 "Gtk couldn't be initialized. "
                 "Use Gtk.init_check() if you want to handle this case.")
@@ -527,7 +513,7 @@ __all__.append('Window')
 
 
 class Dialog(Gtk.Dialog, Container):
-    if GTK2 or GTK3:
+    if GTK3:
         _old_arg_names = ('title', 'parent', 'flags', 'buttons', '_buttons_property')
         _init = deprecated_init(Gtk.Dialog.__init__,
                                 arg_names=('title', 'transient_for', 'flags',
@@ -555,7 +541,7 @@ class Dialog(Gtk.Dialog, Container):
             if add_buttons is not None and not isinstance(add_buttons, Gtk.ButtonsType):
                 warnings.warn('The "buttons" argument must be a Gtk.ButtonsType enum value. '
                               'Please use the "add_buttons" method for adding buttons. '
-                              'See: https://wiki.gnome.org/PyGObject/InitializerDeprecations',
+                              'See: https://wiki.gnome.org/Projects/PyGObject/InitializerDeprecations',
                               PyGTKDeprecationWarning, stacklevel=stacklevel)
                 new_kwargs.pop('buttons', None)
             else:
@@ -565,7 +551,7 @@ class Dialog(Gtk.Dialog, Container):
             if flags:
                 warnings.warn('The "flags" argument for dialog construction is deprecated. '
                               'Please use initializer keywords: modal=True and/or destroy_with_parent=True. '
-                              'See: https://wiki.gnome.org/PyGObject/InitializerDeprecations',
+                              'See: https://wiki.gnome.org/Projects/PyGObject/InitializerDeprecations',
                               PyGTKDeprecationWarning, stacklevel=stacklevel)
 
                 if flags & Gtk.DialogFlags.MODAL:
@@ -581,7 +567,7 @@ class Dialog(Gtk.Dialog, Container):
 
         def run(self, *args, **kwargs):
             with register_sigint_fallback(self.destroy):
-                with wakeup_on_signal():
+                with get_event_loop(GLib.MainContext.default()).running(self.destroy):
                     return Gtk.Dialog.run(self, *args, **kwargs)
 
         action_area = property(lambda dialog: dialog.get_action_area())
@@ -618,7 +604,7 @@ Dialog = override(Dialog)
 __all__.append('Dialog')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class MessageDialog(Gtk.MessageDialog, Dialog):
         __init__ = deprecated_init(Gtk.MessageDialog.__init__,
                                    arg_names=('parent', 'flags', 'message_type',
@@ -639,7 +625,7 @@ if GTK2 or GTK3:
     __all__.append('MessageDialog')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class ColorSelectionDialog(Gtk.ColorSelectionDialog):
         __init__ = deprecated_init(Gtk.ColorSelectionDialog.__init__,
                                    arg_names=('title',),
@@ -657,7 +643,7 @@ if GTK2 or GTK3:
     __all__.append('FileChooserDialog')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class FontSelectionDialog(Gtk.FontSelectionDialog):
         __init__ = deprecated_init(Gtk.FontSelectionDialog.__init__,
                                    arg_names=('title',),
@@ -667,7 +653,7 @@ if GTK2 or GTK3:
     __all__.append('FontSelectionDialog')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class RecentChooserDialog(Gtk.RecentChooserDialog):
         # Note, the "manager" keyword must work across the entire 3.x series because
         # "recent_manager" is not backwards compatible with PyGObject versions prior to 3.10.
@@ -681,7 +667,7 @@ if GTK2 or GTK3:
 
 
 class IconView(Gtk.IconView):
-    if GTK2 or GTK3:
+    if GTK3:
         __init__ = deprecated_init(Gtk.IconView.__init__,
                                    arg_names=('model',),
                                    category=PyGTKDeprecationWarning)
@@ -695,7 +681,7 @@ IconView = override(IconView)
 __all__.append('IconView')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class ToolButton(Gtk.ToolButton):
         __init__ = deprecated_init(Gtk.ToolButton.__init__,
                                    arg_names=('stock_id',),
@@ -815,10 +801,6 @@ class TreeModel(Gtk.TreeModel):
     def __bool__(self):
         return True
 
-    if GTK2 or GTK3:
-        # alias for Python 2.x object protocol
-        __nonzero__ = __bool__
-
     def _getiter(self, key):
         if isinstance(key, Gtk.TreeIter):
             return key
@@ -975,7 +957,7 @@ TreeSortable = override(TreeSortable)
 __all__.append('TreeSortable')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class TreeModelSort(Gtk.TreeModelSort):
         __init__ = deprecated_init(Gtk.TreeModelSort.__init__,
                                    arg_names=('model',),
@@ -995,10 +977,17 @@ class ListStore(Gtk.ListStore, TreeModel, TreeSortable):
         Gtk.ListStore.__init__(self)
         self.set_column_types(column_types)
 
+    # insert_with_valuesv got renamed to insert_with_values with 4.1.0
+    # https://gitlab.gnome.org/GNOME/gtk/-/commit/a1216599ff6b39bca3e9
+    if not hasattr(Gtk.ListStore, "insert_with_valuesv"):
+        insert_with_valuesv = Gtk.ListStore.insert_with_values
+    elif not hasattr(Gtk.ListStore, "insert_with_values"):
+        insert_with_values = Gtk.ListStore.insert_with_valuesv
+
     def _do_insert(self, position, row):
         if row is not None:
             row, columns = self._convert_row(row)
-            treeiter = self.insert_with_valuesv(position, columns, row)
+            treeiter = self.insert_with_values(position, columns, row)
         else:
             treeiter = Gtk.ListStore.insert(self, position)
 
@@ -1194,10 +1183,6 @@ class TreeModelRowIter(object):
         self.iter = self.model.iter_next(self.iter)
         return row
 
-    if GTK2 or GTK3:
-        # alias for Python 2.x object protocol
-        next = __next__
-
     def __iter__(self):
         return self
 
@@ -1343,7 +1328,7 @@ __all__.append('TreeStore')
 
 
 class TreeView(Gtk.TreeView, Container):
-    if GTK2 or GTK3:
+    if GTK3:
         __init__ = deprecated_init(Gtk.TreeView.__init__,
                                    arg_names=('model',),
                                    category=PyGTKDeprecationWarning)
@@ -1352,16 +1337,18 @@ class TreeView(Gtk.TreeView, Container):
     get_visible_range = strip_boolean_result(Gtk.TreeView.get_visible_range)
     get_dest_row_at_pos = strip_boolean_result(Gtk.TreeView.get_dest_row_at_pos)
 
-    def enable_model_drag_source(self, start_button_mask, targets, actions):
-        target_entries = _construct_target_list(targets)
-        super(TreeView, self).enable_model_drag_source(start_button_mask,
-                                                       target_entries,
-                                                       actions)
+    if GTK3:
+        def enable_model_drag_source(self, start_button_mask, targets, actions):
+            target_entries = _construct_target_list(targets)
+            super(TreeView, self).enable_model_drag_source(start_button_mask,
+                                                           target_entries,
+                                                           actions)
 
-    def enable_model_drag_dest(self, targets, actions):
-        target_entries = _construct_target_list(targets)
-        super(TreeView, self).enable_model_drag_dest(target_entries,
-                                                     actions)
+    if GTK3:
+        def enable_model_drag_dest(self, targets, actions):
+            target_entries = _construct_target_list(targets)
+            super(TreeView, self).enable_model_drag_dest(target_entries,
+                                                         actions)
 
     def scroll_to_cell(self, path, column=None, use_align=False, row_align=0.0, col_align=0.0):
         if not isinstance(path, Gtk.TreePath):
@@ -1442,7 +1429,7 @@ TreeSelection = override(TreeSelection)
 __all__.append('TreeSelection')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Button(Gtk.Button, Container):
         _init = deprecated_init(Gtk.Button.__init__,
                                 arg_names=('label', 'stock', 'use_stock', 'use_underline'),
@@ -1501,7 +1488,7 @@ if GTK2 or GTK3:
 
 
 class Adjustment(Gtk.Adjustment):
-    if GTK2 or GTK3:
+    if GTK3:
         _init = deprecated_init(Gtk.Adjustment.__init__,
                                 arg_names=('value', 'lower', 'upper',
                                            'step_increment', 'page_increment', 'page_size'),
@@ -1511,7 +1498,7 @@ class Adjustment(Gtk.Adjustment):
                                 stacklevel=3)
 
     def __init__(self, *args, **kwargs):
-        if GTK2 or GTK3:
+        if GTK3:
             self._init(*args, **kwargs)
             # The value property is set between lower and (upper - page_size).
             # Just in case lower, upper or page_size was still 0 when value
@@ -1534,7 +1521,7 @@ Adjustment = override(Adjustment)
 __all__.append('Adjustment')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Table(Gtk.Table, Container):
         __init__ = deprecated_init(Gtk.Table.__init__,
                                    arg_names=('n_rows', 'n_columns', 'homogeneous'),
@@ -1556,7 +1543,7 @@ if GTK2 or GTK3:
     __all__.append('ScrolledWindow')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class HScrollbar(Gtk.HScrollbar):
         __init__ = deprecated_init(Gtk.HScrollbar.__init__,
                                    arg_names=('adjustment',),
@@ -1574,7 +1561,7 @@ if GTK2 or GTK3:
     __all__.append('VScrollbar')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Paned(Gtk.Paned):
         def pack1(self, child, resize=False, shrink=True):
             super(Paned, self).pack1(child, resize, shrink)
@@ -1586,7 +1573,7 @@ if GTK2 or GTK3:
     __all__.append('Paned')
 
 
-if GTK2 or GTK3:
+if GTK3:
     class Arrow(Gtk.Arrow):
         __init__ = deprecated_init(Gtk.Arrow.__init__,
                                    arg_names=('arrow_type', 'shadow_type'),
@@ -1634,6 +1621,22 @@ class TreeModelFilter(Gtk.TreeModelFilter):
 TreeModelFilter = override(TreeModelFilter)
 __all__.append('TreeModelFilter')
 
+
+if GTK3 or GTK4:
+    class CssProvider(Gtk.CssProvider):
+        def load_from_data(self, text, length=-1):
+            if (Gtk.get_major_version(), Gtk.get_minor_version()) >= (4, 9):
+                if isinstance(text, bytes):
+                    text = text.decode("utf-8")
+                super(CssProvider, self).load_from_data(text, length)
+            else:
+                if isinstance(text, str):
+                    text = text.encode("utf-8")
+                super(CssProvider, self).load_from_data(text)
+
+    CssProvider = override(CssProvider)
+    __all__.append("CssProvider")
+
 if GTK4:
     class CustomSorter(Gtk.CustomSorter):
 
@@ -1664,7 +1667,7 @@ if GTK3:
     Menu = override(Menu)
     __all__.append('Menu')
 
-if GTK2 or GTK3:
+if GTK3:
     _Gtk_main_quit = Gtk.main_quit
 
     @override(Gtk.main_quit)
@@ -1676,15 +1679,30 @@ if GTK2 or GTK3:
     @override(Gtk.main)
     def main(*args, **kwargs):
         with register_sigint_fallback(Gtk.main_quit):
-            with wakeup_on_signal():
+            with get_event_loop(GLib.MainContext.default()).running(Gtk.main_quit):
                 return _Gtk_main(*args, **kwargs)
 
+    _Gtk_main_iteration = Gtk.main_iteration
+    _Gtk_main_iteration_do = Gtk.main_iteration_do
 
-if GTK2 or GTK3:
+    @override(Gtk.main_iteration)
+    def main_iteration():
+        with get_event_loop(GLib.MainContext.default()).paused():
+            return _Gtk_main_iteration()
+
+    @override(Gtk.main_iteration)
+    def main_iteration_do(blocking):
+        with get_event_loop(GLib.MainContext.default()).paused():
+            return _Gtk_main_iteration_do(blocking)
+
+
+if GTK3:
     stock_lookup = strip_boolean_result(Gtk.stock_lookup)
     __all__.append('stock_lookup')
 
-if GTK4:
+if not get_option('legacy_autoinit'):  # TODO: or GTK5
+    initialized = False
+elif GTK4:
     initialized = Gtk.init_check()
 else:
     initialized, argv = Gtk.init_check(sys.argv)
